let appinfo = Services.appinfo;
let options = {
  application: appinfo.ID,
  appversion: appinfo.version,
  platformversion: appinfo.platformVersion,
  os: appinfo.OS,
  osversion: Services.sysinfo.getProperty("version"),
  abi: appinfo.XPCOMABI
};

let man = `
overlay chrome://browser/content/browser.xhtml chrome://keyconfig/content/overlay.xhtml
style   chrome://keyconfig/content/keyconfig.xhtml chrome://keyconfig/skin/keyconfig.css
`;

function showRestartNotifcation(verb, window) {
  window.PopupNotifications._currentNotifications.shift();
  window.PopupNotifications.show(
    window.gBrowser.selectedBrowser,
    'addon-install-restart',
    'Dorando keyconfig has been ' + verb + ', but a restart is required to ' + (verb == 'upgraded' || verb == 're-enabled' ? 'enable' : 'remove') + ' add-on functionality.',
    'addons-notification-icon',
    {
      label: 'Restart Now',
      accessKey: 'R',
      callback() {
        let cancelQuit = Cc['@mozilla.org/supports-PRBool;1'].createInstance(Ci.nsISupportsPRBool);
        Services.obs.notifyObservers(cancelQuit, 'quit-application-requested', 'restart');

        if (cancelQuit.data)
          return;

        if (Services.appinfo.inSafeMode)
          Services.startup.restartInSafeMode(Ci.nsIAppStartup.eAttemptQuit);
        else
          Services.startup.quit(Ci.nsIAppStartup.eAttemptQuit | Ci.nsIAppStartup.eRestart);
      }
    },
    [{
      label: 'Not Now',
      accessKey: 'N',
      callback: () => {},
    }],
    {
      popupIconURL: 'chrome://keyconfig/skin/addon-install-restart.svg',
      persistent: false,
      hideClose: true,
      timeout: Date.now() + 30000,
      removeOnDismissal: true
    }
  );
}

function install() { }

function uninstall() { }

function startup(data, reason) {
  Components.utils.import("chrome://keyconfig/content/defaultPreferencesLoader.jsm");

  function keyconfigService(aTopic) {
   this.os.addObserver(this,aTopic,false);
  }

  keyconfigService.prototype = {
   js: Components.classes["@mozilla.org/moz/jssubscript-loader;1"].getService(Components.interfaces.mozIJSSubScriptLoader),
   os: Components.classes["@mozilla.org/observer-service;1"].getService(Components.interfaces.nsIObserverService),
   ps: Components.classes['@mozilla.org/preferences-service;1']
                 .getService(Components.interfaces.nsIPrefService).getBranch(""),

   observe: function (aSubject, aTopic, aData) {
    if (aTopic == "browser-delayed-startup-finished") {
      // Current Thunderbird nightly builds do not load default preferences
      // from overlay add-ons. They're probably going to fix this, but it may go
      // away again at some point in the future, and in any case we'll need to do
      // it ourselves when we convert from overlay to bootstrapped, and there
      // shouldn't be any harm in setting the default values of preferences twice
      // (i.e., both Thunderbird and our code doing it).
      // This is in a try/catch because if it fails it's probably because
      // setStringPref failed, in which case we're running inside an earlier
      // application version which has already loaded the default preferences
      // automatically.
      try {
          var loader = new DefaultPreferencesLoader();
          loader.parseUri(
              "chrome://keyconfig-defaults/content/preferences/keyconfig.js");
      } catch (ex) {}
      aSubject.keyconfig = {service: this};
      aSubject.keyconfig.service.init.call(aSubject);
      this.os.removeObserver(this,"browser-delayed-startup-finished");
      this.os.addObserver(this,"domwindowopened",false);
    } else if (aTopic == "domwindowopened") {
      aSubject.keyconfig = {service: this};
      aSubject.addEventListener("DOMContentLoaded", function () {
        aSubject.keyconfig.service.init.call(aSubject);
      }, { once: true });
    }
   },

   init: function() {
    this.keyconfig.removedKeys = this.document.documentElement.appendChild(this.document.createXULElement("keyconfig"));

   
    if (this.keyconfig.service.ps.prefHasUserValue("keyconfig.global.20110522")) {
      var oldBranch = "keyconfig."; var newBranch = "extensions.dorandoKeyConfig.";
      var oldKeys = this.keyconfig.service.ps.getChildList(oldBranch)
      for (var i = 0; i < oldKeys.length; i++) {
        var newKey = newBranch + oldKeys[i].split(oldBranch)[1];
        switch (this.keyconfig.service.ps.getPrefType(oldKeys[i])) {
          case 32: //PREF_STRING
            var value = this.keyconfig.service.ps.getCharPref(oldKeys[i]);
            this.keyconfig.service.ps.setCharPref(newKey, value);
            break;

          case 64: //PREF_INT
            var value = this.keyconfig.service.ps.getIntPref(oldKeys[i]);
            this.keyconfig.service.ps.setIntPref(newKey, value);
            break;
          
          case 128: //PREF_BOOLEAN
            var value = this.keyconfig.service.ps.getBoolPref(oldKeys[i]);
            this.keyconfig.service.ps.setBoolPref(newKey, value);
            break;
        }
      }
      this.keyconfig.service.ps.deleteBranch(oldBranch); 
    }
    
    this.keyconfig.service.ps.deleteBranch("extensions.dorandoKeyConfig.global")
    
    this.keyconfig.profile = "extensions.dorandoKeyConfig." + this.keyconfig.service.ps.getCharPref("extensions.dorandoKeyConfig.profile") + ".";

    var i, l;

    var keyset = this.document.getElementsByTagName("keyset")[0] ||
                 this.document.documentElement.appendChild(this.document.createXULElement("keyset"));

    var nodes = this.document.getElementsByTagName("key");
    for(i = 0, l = nodes.length; i < l; i++) if(!nodes[i].id)
     nodes[i].id = "xxx_key"+ i +"_"+nodes[i].getAttribute("command")+nodes[i].getAttribute("oncommand");

    var keys = this.keyconfig.service.ps.getChildList(this.keyconfig.profile, {});

    for(i = 0, l = keys.length; i < l; i++) {
     var key, node;
     try {
         try {
             // Gecko 58+
             key = this.keyconfig.service.ps.getStringPref(keys[i]).split("][");
         }
         catch (e) {
             key = this.keyconfig.service.ps.getComplexValue(keys[i], Components.interfaces.nsISupportsString).data.split("][");
         }
     } catch(e) { continue; }
     if(key[3] && (!key[4] || key[4] == this.document.location)) {
      node = keyset.appendChild(this.document.createXULElement("key"));
      node.id = keys[i].substr(this.keyconfig.profile.length);
  //    node.addEventListener("command",key[3]);
      node.setAttribute("oncommand",key[3])
     } else {
      node = this.document.getElementById(keys[i].substr(this.keyconfig.profile.length));
      if(!node) continue;
     }

     node.removeAttribute("modifiers"); node.removeAttribute("keycode");
     node.removeAttribute("charcode"); node.removeAttribute("keytext");
     if(key[0] == "!") {this.keyconfig.removedKeys.appendChild(node); continue;}

     if(key[0]) node.setAttribute("modifiers",key[0]);
     if(key[1]) node.setAttribute("key",key[1]);
     if(key[2]) node.setAttribute("keycode",key[2]);
    }
   }

  }

  Components.utils.import("chrome://keyconfig/content/ChromeManifest.jsm");
  Components.utils.import("chrome://keyconfig/content/Overlays.jsm");
  const { AppConstants } = ChromeUtils.import('resource://gre/modules/AppConstants.jsm');

  const window = Services.wm.getMostRecentWindow('navigator:browser');
  var isXul = AppConstants.BROWSER_CHROME_URL.endsWith('.xul');
  if ((reason === ADDON_UPGRADE || reason === ADDON_DOWNGRADE || reason === ADDON_INSTALL || reason === ADDON_ENABLE || reason === APP_STARTUP)
      && (!Services.prefs.prefHasUserValue('extensions.dorandoKeyConfig.migrationXHTML')
          || ((Services.prefs.getBoolPref('extensions.dorandoKeyConfig.migrationXHTML') && isXul)
              || (!Services.prefs.getBoolPref('extensions.dorandoKeyConfig.migrationXHTML') && !isXul)))) {
    var url = isXul ? ['xhtml', 'xul'] : ['xul', 'xhtml'];
    Services.prefs.getChildList('extensions.dorandoKeyConfig.').forEach(pref => {
      if (Services.prefs.getPrefType(pref) == 32 && Services.prefs.getStringPref(pref).endsWith('][chrome://browser/content/browser.' + url[0])) {
        Services.prefs.setStringPref(pref, Services.prefs.getStringPref(pref).replace('][chrome://browser/content/browser.' + url[0], '][chrome://browser/content/browser.' + url[1]));
      }
    });
    Services.prefs.setBoolPref('extensions.dorandoKeyConfig.migrationXHTML', !isXul);
  }
  if (reason === ADDON_UPGRADE || reason === ADDON_DOWNGRADE) {
      showRestartNotifcation("upgraded", window);
      return;
  } else if (reason === ADDON_ENABLE && window.keyconfig) {
      showRestartNotifcation("re-enabled", window);
      return;
  }

  if (reason === ADDON_INSTALL || (reason === ADDON_ENABLE && !window.keyconfig)) {
    var kS = new keyconfigService("domwindowopened");
    try {
        var loader = new DefaultPreferencesLoader();
        loader.parseUri(
            "chrome://keyconfig-defaults/content/preferences/keyconfig.js");
    } catch (ex) {}
    var enumerator = Services.wm.getEnumerator(null);
    while (enumerator.hasMoreElements()) {
      var win = enumerator.getNext();

      win.keyconfig = {service: kS};
      win.keyconfig.service.init.call(win);
      var keyset = win.document.getElementsByTagName("keyset").length && win.document.getElementsByTagName("keyset")[0];
      if (keyset)
        keyset.parentNode.insertBefore(keyset, keyset.nextSibling);

      (async function (win) {
        let chromeManifest = new ChromeManifest(function () { return man; }, options);
        await chromeManifest.parse();
        if (win.document.createXULElement) {
          Overlays.load(chromeManifest, win.document.defaultView);
        }
      })(win);
    }
  } else {
    new keyconfigService("browser-delayed-startup-finished");
  }

  (async function () {
    let chromeManifest = new ChromeManifest(function () { return man; }, options);
    await chromeManifest.parse();

    let documentObserver = {
      observe(document) {
        if (document.createXULElement) {
          Overlays.load(chromeManifest, document.defaultView);
        }
      }
    };
    Services.obs.addObserver(documentObserver, "chrome-document-loaded");
  })();
}

function shutdown(data, reason) {
  const window = Services.wm.getMostRecentWindow('navigator:browser');
  if (reason === ADDON_DISABLE) {
      showRestartNotifcation("disabled", window);
      return;
  } else if (reason === ADDON_UNINSTALL && window.keyconfig) {
      showRestartNotifcation("uninstalled", window);
      return;
  }
}